from base64 import b64encode
from collections.abc import AsyncGenerator, Sequence
from typing import Any, Literal, cast, overload
from uuid import uuid4

from draive.gemini.client import GeminiClient
from draive.gemini.config import GeminiConfig
from draive.gemini.errors import GeminiException
from draive.gemini.models import (
    GeminiDataMessageContent,
    GeminiDataReferenceMessageContent,
    GeminiFunctionCallMessageContent,
    GeminiFunctionsTool,
    GeminiFunctionToolSpecification,
    GeminiGenerationResult,
    GeminiMessage,
    GeminiMessageContent,
    GeminiRequestMessage,
    GeminiTextMessageContent,
)
from draive.metrics import ArgumentsTrace, ResultTrace, TokenUsage
from draive.parameters import DataModel, ToolSpecification
from draive.scope import ctx
from draive.types import (
    AudioBase64Content,
    AudioDataContent,
    AudioURLContent,
    ImageBase64Content,
    ImageDataContent,
    ImageURLContent,
    LMMCompletion,
    LMMCompletionChunk,
    LMMContextElement,
    LMMInput,
    LMMInstruction,
    LMMOutput,
    LMMOutputStream,
    LMMOutputStreamChunk,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
    MultimodalContentElement,
    TextContent,
    VideoBase64Content,
    VideoDataContent,
    VideoURLContent,
)

__all__ = [
    "gemini_lmm_invocation",
]


@overload
async def gemini_lmm_invocation(
    *,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    require_tool: ToolSpecification | bool = False,
    output: Literal["text", "json"] = "text",
    stream: Literal[True],
    **extra: Any,
) -> LMMOutputStream: ...


@overload
async def gemini_lmm_invocation(
    *,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    require_tool: ToolSpecification | bool = False,
    output: Literal["text", "json"] = "text",
    stream: Literal[False] = False,
    **extra: Any,
) -> LMMOutput: ...


@overload
async def gemini_lmm_invocation(
    *,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    require_tool: ToolSpecification | bool = False,
    output: Literal["text", "json"] = "text",
    stream: bool = False,
    **extra: Any,
) -> LMMOutputStream | LMMOutput: ...


async def gemini_lmm_invocation(
    *,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    require_tool: ToolSpecification | bool = False,
    output: Literal["text", "json"] = "text",
    stream: bool = False,
    **extra: Any,
) -> LMMOutputStream | LMMOutput:
    with ctx.nested(
        "gemini_lmm_completion",
        metrics=[
            ArgumentsTrace.of(
                context=context,
                tools=tools,
                require_tool=require_tool,
                output=output,
                stream=stream,
                **extra,
            ),
        ],
    ):
        ctx.log_debug("Requested Gemini lmm")
        client: GeminiClient = ctx.dependency(GeminiClient)
        config: GeminiConfig = ctx.state(GeminiConfig).updated(**extra)
        ctx.record(config)

        match output:
            case "text":
                config = config.updated(response_format="text/plain")

            case "json":
                if tools:
                    ctx.log_warning(
                        "Attempting to use Gemini in JSON mode with tools which is not supported."
                        " Using text mode instead..."
                    )
                    config = config.updated(response_format="text/plain")

                else:
                    config = config.updated(response_format="application/json")

        instruction: str = ""
        messages: list[GeminiRequestMessage] = []
        for element in context:
            match element:
                case LMMInstruction() as instruction_element:
                    instruction += instruction_element.content

                case other:
                    messages.append(_convert_context_element(element=other))

        if stream:
            return ctx.stream(
                generator=_generation_stream(
                    client=client,
                    config=config,
                    instruction=instruction,
                    messages=messages,
                    tools=tools,
                    require_tool=require_tool,
                ),
            )

        else:
            return await _generate(
                client=client,
                config=config,
                instruction=instruction,
                messages=messages,
                tools=tools,
                require_tool=require_tool,
            )


def _convert_content_element(  # noqa: C901, PLR0911
    element: MultimodalContentElement,
) -> dict[str, Any]:
    match element:
        case TextContent() as text:
            return {"text": text.text}

        case ImageURLContent() as image:
            return {
                "fileData": {
                    "mimeType": image.mime_type or "image",
                    "fileUri": image.image_url,
                }
            }

        case ImageBase64Content() as image:
            return {
                "inlineData": {
                    "mimeType": image.mime_type or "image",
                    "data": image.image_base64,
                }
            }

        case ImageDataContent() as image:
            return {
                "inlineData": {
                    "mimeType": image.mime_type or "image",
                    "data": b64encode(image.image_data).decode("utf-8"),
                }
            }

        case AudioURLContent() as audio:
            return {
                "fileData": {
                    "mimeType": audio.mime_type or "audio",
                    "fileUri": audio.audio_url,
                }
            }

        case AudioBase64Content() as audio:
            return {
                "inlineData": {
                    "mimeType": audio.mime_type or "audio",
                    "data": audio.audio_base64,
                }
            }

        case AudioDataContent() as audio:
            return {
                "inlineData": {
                    "mimeType": audio.mime_type or "audio",
                    "data": b64encode(audio.audio_data).decode("utf-8"),
                }
            }

        case VideoURLContent() as video:
            return {
                "fileData": {
                    "mimeType": video.mime_type or "video",
                    "fileUri": video.video_url,
                }
            }

        case VideoBase64Content() as video:
            return {
                "inlineData": {
                    "mimeType": video.mime_type or "video",
                    "data": video.video_base64,
                }
            }

        case VideoDataContent() as video:
            return {
                "inlineData": {
                    "mimeType": video.mime_type or "video",
                    "data": b64encode(video.video_data).decode("utf-8"),
                }
            }

        case DataModel() as data:
            return {"text": data.as_json()}


def _convert_context_element(
    element: LMMContextElement,
) -> GeminiRequestMessage:
    match element:
        case LMMInstruction():
            raise ValueError("Instruction has to be processed separately")

        case LMMInput() as input:
            return {
                "role": "user",
                "parts": [
                    _convert_content_element(element=element) for element in input.content.parts
                ],
            }

        case LMMCompletion() as completion:
            return {
                "role": "model",
                "parts": [
                    _convert_content_element(element=element)
                    for element in completion.content.parts
                ],
            }

        case LMMToolRequests() as tool_requests:
            return {
                "role": "model",
                "parts": [
                    {
                        "functionCall": {
                            "name": request.tool,
                            "args": request.arguments,
                        },
                    }
                    for request in tool_requests.requests
                ],
            }

        case LMMToolResponse() as tool_response:
            return {
                "role": "model",
                "parts": [
                    {
                        "functionResponse": {
                            "name": tool_response.tool,
                            "response": tool_response.content.as_dict(),
                        },
                    }
                ],
            }


def _convert_content_part(  # noqa: PLR0911
    part: GeminiMessageContent,
) -> MultimodalContentElement:
    match part:
        case GeminiTextMessageContent() as text:
            return TextContent(text=text.text)

        case GeminiDataMessageContent() as data:
            mime_type: str = data.data.mime_type
            if mime_type.startswith("image"):
                return ImageBase64Content(
                    mime_type=mime_type,
                    image_base64=data.data.data,
                )

            elif mime_type.startswith("audio"):
                return AudioBase64Content(
                    mime_type=mime_type,
                    audio_base64=data.data.data,
                )

            elif mime_type.startswith("video"):
                return VideoBase64Content(
                    mime_type=mime_type,
                    video_base64=data.data.data,
                )

            else:
                raise GeminiException("Unsupported result content data %s", data)

        case GeminiDataReferenceMessageContent() as reference:
            mime_type: str = reference.reference.mime_type
            if mime_type.startswith("image"):
                return ImageURLContent(
                    mime_type=mime_type,
                    image_url=reference.reference.uri,
                )

            elif mime_type.startswith("audio"):
                return AudioURLContent(
                    mime_type=mime_type,
                    audio_url=reference.reference.uri,
                )

            elif mime_type.startswith("video"):
                return VideoURLContent(
                    mime_type=mime_type,
                    video_url=reference.reference.uri,
                )

            else:
                raise GeminiException("Unsupported result content reference %s", reference)

        case other:
            raise GeminiException("Unsupported result content %s", other)


async def _generate(  # noqa: PLR0913, C901, PLR0912
    *,
    client: GeminiClient,
    config: GeminiConfig,
    instruction: str,
    messages: list[GeminiRequestMessage],
    tools: Sequence[ToolSpecification] | None,
    require_tool: ToolSpecification | bool,
) -> LMMOutput:
    result: GeminiGenerationResult
    match require_tool:
        case bool(required):
            result = await client.generate(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[
                    GeminiFunctionsTool(
                        functionDeclarations=[
                            # those models are the same, can safely cast
                            cast(GeminiFunctionToolSpecification, tool["function"])
                        ]
                    )
                    for tool in tools or []
                ],
                suggest_tools=required,
            )

        case tool:
            assert tool in (tools or []), "Can't suggest a tool without using it"  # nosec: B101
            result = await client.generate(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[
                    GeminiFunctionsTool(
                        functionDeclarations=[
                            # those models are the same, can safely cast
                            cast(GeminiFunctionToolSpecification, tool["function"])
                        ]
                    )
                    for tool in tools or []
                ],
                suggest_tools=True,
            )

    if usage := result.usage:
        ctx.record(
            TokenUsage.for_model(
                config.model,
                input_tokens=usage.prompt_tokens,
                output_tokens=usage.generated_tokens,
            ),
        )

    if not result.choices:
        raise GeminiException("Invalid Gemini completion - missing messages!", result)

    generated_message: GeminiMessage = result.choices[0].content

    message_parts: list[
        GeminiTextMessageContent | GeminiDataReferenceMessageContent | GeminiDataMessageContent
    ] = []
    tool_calls: list[GeminiFunctionCallMessageContent] = []
    for part in generated_message.content:
        match part:
            case GeminiTextMessageContent() as text:
                message_parts.append(text)

            case GeminiFunctionCallMessageContent() as call:
                tool_calls.append(call)

            case GeminiDataReferenceMessageContent() as reference:
                message_parts.append(reference)

            case GeminiDataMessageContent() as data:
                message_parts.append(data)

            case other:
                raise GeminiException("Invalid Gemini completion part", other)

    if tool_calls and (tools := tools):
        assert not message_parts, "Unexpected content when calling tools"  # nosec: B101
        ctx.record(ResultTrace.of(tool_calls))

        return LMMToolRequests(
            requests=[
                LMMToolRequest(
                    identifier=uuid4().hex,
                    tool=call.function_call.name,
                    arguments=call.function_call.arguments,
                )
                for call in tool_calls
            ]
        )

    elif message_parts:
        ctx.record(ResultTrace.of(message_parts))
        return LMMCompletion.of(*[_convert_content_part(part) for part in message_parts])

    else:
        raise GeminiException("Invalid Gemini completion", result)


async def _generation_stream(  # noqa: PLR0913
    *,
    client: GeminiClient,
    config: GeminiConfig,
    instruction: str,
    messages: list[GeminiRequestMessage],
    tools: Sequence[ToolSpecification] | None,
    require_tool: ToolSpecification | bool,
) -> AsyncGenerator[LMMOutputStreamChunk, None]:
    ctx.log_warning("Gemini streaming api is not supported yet, using regular response...")
    output: LMMOutput = await _generate(
        client=client,
        config=config,
        instruction=instruction,
        messages=messages,
        tools=tools,
        require_tool=require_tool,
    )

    match output:
        case LMMCompletion() as completion:
            yield LMMCompletionChunk.of(completion.content)

        case other:
            yield other



from asyncio import AbstractEventLoop, CancelledError, Event, Future, get_running_loop
from collections import deque
from collections.abc import AsyncIterator
from typing import Self

__all__ = [
    "AsyncStream",
]


class AsyncStream[Element](AsyncIterator[Element]):
    def __init__(
        self,
        loop: AbstractEventLoop | None = None,
    ) -> None:
        self._loop: AbstractEventLoop = loop or get_running_loop()
        self._buffer: deque[Element] = deque()
        self._waiting_queue: deque[Future[Element]] = deque()
        self._finish_exception: BaseException | None = None
        self._finished: Event = Event()

    def __del__(self) -> None:
        while self._waiting_queue:
            waiting: Future[Element] = self._waiting_queue.popleft()
            if waiting.done():
                continue

            else:
                waiting.set_exception(CancelledError())

        self._finished.set()

    @property
    def finished(self) -> bool:
        return self._finished.is_set()

    async def wait(self) -> None:
        await self._finished.wait()

    def send(
        self,
        element: Element,
        /,
        *elements: Element,
    ) -> None:
        if self.finished:
            raise RuntimeError("AsyncStream has been already finished")

        while self._waiting_queue:
            assert not self._buffer  # nosec: B101
            waiting: Future[Element] = self._waiting_queue.popleft()
            if waiting.done():
                continue
            else:
                waiting.set_result(element)
                break
        else:
            self._buffer.append(element)

        self._buffer.extend(elements)

    def finish(
        self,
        exception: BaseException | None = None,
    ) -> None:
        if self.finished:
            raise RuntimeError("AsyncStream has been already finished")

        self._finish_exception = exception or StopAsyncIteration()
        self._finished.set()

        if self._buffer:
            return  # allow consuming buffer to the end

        while self._waiting_queue:
            waiting: Future[Element] = self._waiting_queue.popleft()
            if waiting.done():
                continue
            else:
                waiting.set_exception(self._finish_exception)

    def __aiter__(self) -> Self:
        return self

    async def __anext__(self) -> Element:
        if self._buffer:  # use buffer first
            return self._buffer.popleft()
        if finish_exception := self._finish_exception:  # check if finished
            raise finish_exception

        # create new waiting future
        future: Future[Element] = self._loop.create_future()
        self._waiting_queue.append(future)

        # wait for the result
        return await future



from asyncio import FIRST_COMPLETED, Task, gather, wait
from collections.abc import AsyncIterator
from typing import Any, Literal, final

from draive.lmm.errors import ToolException
from draive.lmm.state import ToolStatusStream
from draive.lmm.tool import AnyTool
from draive.parameters import ToolSpecification
from draive.scope import ctx
from draive.types import (
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
    MultimodalContent,
    ToolCallStatus,
)
from draive.utils import AsyncStream, freeze

__all__ = [
    "Toolbox",
]


@final
class Toolbox:
    def __init__(
        self,
        *tools: AnyTool,
        suggest: AnyTool | Literal[True] | None = None,
        recursive_calls_limit: int | None = None,
    ) -> None:
        self._tools: dict[str, AnyTool] = {tool.name: tool for tool in tools}
        self.recursion_limit: int = recursive_calls_limit or 1
        self.suggest_tools: bool
        self._suggested_tool: AnyTool | None
        match suggest:
            case None:
                self.suggest_tools = False
                self._suggested_tool = None
            case True:
                self.suggest_tools = True if self._tools else False
                self._suggested_tool = None
            case tool:
                self.suggest_tools = True
                self._suggested_tool = tool
                self._tools[tool.name] = tool

        freeze(self)

    @property
    def call_range(self) -> range:
        return range(0, self.recursion_limit + 1)

    def tool_suggestion(
        self,
        recursion_level: int = 0,
    ) -> ToolSpecification | bool:
        if recursion_level != 0:
            return False  # suggest tools only for the first call

        elif self._suggested_tool is not None:
            return self._suggested_tool.specification if self._suggested_tool.available else False

        else:
            return self.suggest_tools

    def available_tools(
        self,
        recursion_level: int = 0,
    ) -> list[ToolSpecification]:
        if recursion_level <= self.recursion_limit:
            return [tool.specification for tool in self._tools.values() if tool.available]
        else:
            ctx.log_warning("Reached tool calls recursion limit, ignoring tools...")
            return []  # disable tools when reached recursion limit

    async def call_tool(
        self,
        name: str,
        /,
        call_id: str,
        arguments: dict[str, Any],
    ) -> MultimodalContent:
        if tool := self._tools.get(name):
            return await tool._toolbox_call(  # pyright: ignore[reportPrivateUsage]
                call_id,
                arguments=arguments,
            )

        else:
            raise ToolException("Requested tool is not defined", name)

    async def respond(
        self,
        requests: LMMToolRequests,
        /,
    ) -> list[LMMToolResponse]:
        return await gather(
            *[self._respond(request) for request in requests.requests],
            return_exceptions=False,
        )

    async def _respond(
        self,
        request: LMMToolRequest,
        /,
    ) -> LMMToolResponse:
        if tool := self._tools.get(request.tool):
            return LMMToolResponse(
                identifier=request.identifier,
                tool=request.tool,
                content=await tool._toolbox_call(  # pyright: ignore[reportPrivateUsage]
                    request.identifier,
                    arguments=request.arguments or {},
                ),
                direct=tool.requires_direct_result,
            )

        else:
            # log error and provide fallback result to avoid blowing out the execution
            ctx.log_error("Requested tool (%s) is not defined", request.tool)
            return LMMToolResponse(
                identifier=request.identifier,
                tool=request.tool,
                content=MultimodalContent.of("ERROR"),
                direct=False,
            )

    def stream(
        self,
        requests: LMMToolRequests,
        /,
    ) -> AsyncIterator[LMMToolResponse | ToolCallStatus]:
        stream: AsyncStream[LMMToolResponse | ToolCallStatus] = AsyncStream()

        async def tools_stream() -> None:
            try:
                with ctx.updated(ToolStatusStream(send=stream.send)):
                    pending_tasks: set[Task[LMMToolResponse]] = {
                        ctx.spawn_subtask(
                            self._respond,
                            request,
                        )
                        for request in requests.requests
                    }

                    while pending_tasks:
                        done, pending_tasks = await wait(
                            pending_tasks,
                            return_when=FIRST_COMPLETED,
                        )
                        for task in done:
                            stream.send(task.result())

            except BaseException as exc:
                stream.finish(exception=exc)

            else:
                stream.finish()

        ctx.spawn_subtask(tools_stream)
        return stream



from collections.abc import Iterable
from typing import Any

from draive.generation.text.state import TextGeneration
from draive.lmm import Toolbox
from draive.scope import ctx
from draive.types import Instruction, MultimodalContent, MultimodalContentConvertible

__all__ = [
    "generate_text",
]


async def generate_text(
    *,
    instruction: Instruction | str,
    input: MultimodalContent | MultimodalContentConvertible,  # noqa: A002
    tools: Toolbox | None = None,
    examples: Iterable[tuple[MultimodalContent | MultimodalContentConvertible, str]] | None = None,
    **extra: Any,
) -> str:
    return await ctx.state(TextGeneration).generate(
        instruction=instruction,
        input=input,
        tools=tools,
        examples=examples,
        **extra,
    )



from asyncio import Task, TaskGroup, current_task, shield
from collections.abc import AsyncGenerator, AsyncIterator, Callable, Coroutine, Iterable
from contextvars import ContextVar, Token
from logging import Logger, getLogger
from types import TracebackType
from typing import Any, final

from draive.metrics import (
    ExceptionTrace,
    Metric,
    MetricsTrace,
    MetricsTraceReporter,
    metrics_log_reporter,
)
from draive.parameters import ParametrizedData
from draive.scope.dependencies import ScopeDependencies, ScopeDependency
from draive.scope.errors import MissingScopeContext
from draive.scope.state import ScopeState
from draive.utils import AsyncStream, getenv_bool, mimic_function

__all__ = [
    "ctx",
]

_TaskGroup_Var = ContextVar[TaskGroup]("_TaskGroup_Var")
_MetricsScope_Var = ContextVar[MetricsTrace]("_MetricsScope_Var")
_StateScope_Var = ContextVar[ScopeState]("_ScopeState_Var")
_DependenciesScope_Var = ContextVar[ScopeDependencies]("_DependenciesScope_Var")


class _RootContext:
    def __init__(  # noqa: PLR0913
        self,
        task_group: TaskGroup,
        dependencies: ScopeDependencies,
        state: ScopeState,
        metrics: MetricsTrace,
        trace_reporting: MetricsTraceReporter | None,
    ) -> None:
        self._task_group: TaskGroup = task_group
        self._task_group_token: Token[TaskGroup] | None = None
        self._dependencies: ScopeDependencies = dependencies
        self._dependencies_token: Token[ScopeDependencies] | None = None
        self._state: ScopeState = state
        self._state_token: Token[ScopeState] | None = None
        self._metrics: MetricsTrace = metrics
        self._metrics_token: Token[MetricsTrace] | None = None
        self._report_trace: MetricsTraceReporter | None = trace_reporting

    async def __aenter__(self) -> None:
        # start the task group first
        assert self._task_group_token is None, "Reentrance is not allowed"  # nosec: B101
        self._task_group_token = _TaskGroup_Var.set(self._task_group)
        await self._task_group.__aenter__()
        # prepare state scope
        assert self._state_token is None, "Reentrance is not allowed"  # nosec: B101
        self._state_token = _StateScope_Var.set(self._state)
        # then initialize dependencies
        assert self._dependencies_token is None, "Reentrance is not allowed"  # nosec: B101
        self._dependencies_token = _DependenciesScope_Var.set(self._dependencies)
        # finally begin metrics capture
        assert self._metrics_token is None, "Reentrance is not allowed"  # nosec: B101
        self._metrics.enter()
        self._metrics_token = _MetricsScope_Var.set(self._metrics)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        assert self._task_group_token is not None, "Can't exit scope without entering"  # nosec: B101
        _TaskGroup_Var.reset(self._task_group_token)
        # finish task group - wait for completion
        try:
            await self._task_group.__aexit__(
                et=exc_type,
                exc=exc_val,
                tb=exc_tb,
            )

        except BaseException as exc:
            # record task group exceptions
            self._metrics.record(ExceptionTrace.of(exc_val or exc))

        else:
            # or context exception
            if (exception := exc_val) and exc_type is not GeneratorExit:
                self._metrics.record(ExceptionTrace.of(exception))

        finally:
            # then end metrics capture
            assert self._metrics_token is not None, "Can't exit scope without entering"  # nosec: B101
            _MetricsScope_Var.reset(self._metrics_token)
            self._metrics.exit()
            assert (  # nosec: B101
                self._metrics.is_finished
            ), "Unbalanced metrics trace enter/exit calls, possibly an unstructured task running"
            # report metrics trace
            if report_trace := self._report_trace:
                # it still have access to the dependencies and state
                try:  # catch all exceptions - we don't want to blow up on metrics
                    await shield(
                        report_trace(
                            trace_id=self._metrics.trace_id,
                            logger=self._metrics._logger,  # pyright: ignore[reportPrivateUsage]
                            report=self._metrics.report(),
                        )
                    )

                except Exception as exc:
                    self._metrics.log_error(
                        "Failed to finish metrics trace report",
                        exception=exc,
                    )
                    return None

            # cleanup dependencies next
            assert self._dependencies_token is not None, "Can't exit scope without entering"  # nosec: B101
            _DependenciesScope_Var.reset(self._dependencies_token)
            # finally reset state
            assert self._state_token is not None, "Can't exit scope without entering"  # nosec: B101
            _StateScope_Var.reset(self._state_token)


class _PartialContext:
    def __init__(
        self,
        task_group: TaskGroup | None = None,
        metrics: MetricsTrace | None = None,
        state: ScopeState | None = None,
    ) -> None:
        self._task_group: TaskGroup | None = task_group
        self._task_group_token: Token[TaskGroup] | None = None
        self._metrics: MetricsTrace | None = metrics
        self._metrics_token: Token[MetricsTrace] | None = None
        self._state: ScopeState | None = state
        self._state_token: Token[ScopeState] | None = None

    def __enter__(self) -> None:
        if metrics := self._metrics:
            assert self._metrics_token is None, "Reentrance is not allowed"  # nosec: B101
            metrics.enter()
            self._metrics_token = _MetricsScope_Var.set(metrics)
        if state := self._state:
            assert self._state_token is None, "Reentrance is not allowed"  # nosec: B101
            self._state_token = _StateScope_Var.set(state)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if (token := self._metrics_token) and (metrics := self._metrics):
            _MetricsScope_Var.reset(self._metrics_token)
            if (exception := exc_val) and exc_type is not GeneratorExit:
                metrics.record(ExceptionTrace.of(exception))

            metrics.exit()

        if token := self._state_token:
            _StateScope_Var.reset(token)

    async def __aenter__(self) -> None:
        if task_group := self._task_group:
            assert self._task_group_token is None, "Reentrance is not allowed"  # nosec: B101
            self._task_group_token = _TaskGroup_Var.set(task_group)
            await task_group.__aenter__()

        self.__enter__()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        try:
            if task_group := self._task_group:
                assert self._task_group_token is not None, "Can't exit scope without entering"  # nosec: B101
                _TaskGroup_Var.reset(self._task_group_token)
                await task_group.__aexit__(
                    et=exc_type,
                    exc=exc_val,
                    tb=exc_tb,
                )

        except BaseException as exc:
            self.__exit__(
                exc_type=type(exc),
                exc_val=exc,
                exc_tb=exc.__traceback__,
            )

        else:
            self.__exit__(
                exc_type=exc_type,
                exc_val=exc_val,
                exc_tb=exc_tb,
            )


@final
class ctx:
    @staticmethod
    def new(  # noqa: PLR0913
        label: str | None = None,
        *,
        dependencies: ScopeDependencies
        | Iterable[type[ScopeDependency] | ScopeDependency]
        | None = None,
        state: ScopeState | Iterable[ParametrizedData] | None = None,
        metrics: Iterable[Metric] | None = None,
        logger: Logger | None = None,
        trace_reporting: MetricsTraceReporter | None = None,
    ) -> _RootContext:
        root_dependencies: ScopeDependencies
        if dependencies is None:
            root_dependencies = ScopeDependencies()
        elif isinstance(dependencies, ScopeDependencies):
            root_dependencies = dependencies
        else:
            root_dependencies = ScopeDependencies(*dependencies)

        root_state: ScopeState
        if state is None:
            root_state = ScopeState()
        elif isinstance(state, ScopeState):
            root_state = state
        else:
            root_state = ScopeState(*state)

        root_logger: Logger = logger or getLogger(name=label)

        trace_reporter: MetricsTraceReporter | None
        if trace_reporting:
            trace_reporter = trace_reporting
        elif getenv_bool("DEBUG_LOGGING", __debug__):
            trace_reporter = metrics_log_reporter(
                list_items_limit=-4,
                item_character_limit=64,
            )
        else:
            trace_reporter = None

        return _RootContext(
            task_group=TaskGroup(),
            dependencies=root_dependencies,
            state=root_state,
            metrics=MetricsTrace(
                label=label,
                logger=root_logger,
                parent=None,
                metrics=metrics,
            ),
            trace_reporting=trace_reporter,
        )

    @staticmethod
    def wrap[**Args, Result](  # noqa: PLR0913
        label: str | None = None,
        *,
        dependencies: ScopeDependencies
        | Iterable[type[ScopeDependency] | ScopeDependency]
        | None = None,
        state: ScopeState | Iterable[ParametrizedData] | None = None,
        metrics: Iterable[Metric] | None = None,
        logger: Logger | None = None,
        trace_reporting: MetricsTraceReporter | None = None,
    ) -> Callable[
        [Callable[Args, Coroutine[None, None, Result]]],
        Callable[Args, Coroutine[None, None, Result]],
    ]:
        root_dependencies: ScopeDependencies
        if dependencies is None:
            root_dependencies = ScopeDependencies()
        elif isinstance(dependencies, ScopeDependencies):
            root_dependencies = dependencies
        else:
            root_dependencies = ScopeDependencies(*dependencies)

        root_state: ScopeState
        if state is None:
            root_state = ScopeState()
        elif isinstance(state, ScopeState):
            root_state = state
        else:
            root_state = ScopeState(*state)

        trace_reporter: MetricsTraceReporter | None
        if trace_reporting:
            trace_reporter = trace_reporting
        elif getenv_bool("DEBUG_LOGGING", __debug__):
            trace_reporter = metrics_log_reporter(
                list_items_limit=-4,
                item_character_limit=64,
            )
        else:
            trace_reporter = None

        def wrapper(
            function: Callable[Args, Coroutine[None, None, Result]],
            /,
        ) -> Callable[Args, Coroutine[None, None, Result]]:
            @mimic_function(function)
            async def wrapped(*args: Args.args, **kwargs: Args.kwargs) -> Result:
                async with ctx.new(
                    label,
                    dependencies=root_dependencies,
                    state=root_state,
                    metrics=metrics,
                    logger=logger,
                    trace_reporting=trace_reporter,
                ):
                    return await function(*args, **kwargs)

            return wrapped

        return wrapper

    @staticmethod
    def update[**Args, Result](
        *state: ParametrizedData,
    ) -> Callable[
        [Callable[Args, Coroutine[None, None, Result]]],
        Callable[Args, Coroutine[None, None, Result]],
    ]:
        def wrapper(
            function: Callable[Args, Coroutine[None, None, Result]],
            /,
        ) -> Callable[Args, Coroutine[None, None, Result]]:
            @mimic_function(function)
            async def wrapped(*args: Args.args, **kwargs: Args.kwargs) -> Result:
                with ctx.updated(*state):
                    return await function(*args, **kwargs)

            return wrapped

        return wrapper

    @staticmethod
    def _current_task_group() -> TaskGroup:
        try:
            return _TaskGroup_Var.get()
        except LookupError as exc:
            raise MissingScopeContext("TaskGroup requested but not defined!") from exc

    @staticmethod
    def _current_metrics() -> MetricsTrace:
        try:
            return _MetricsScope_Var.get()
        except LookupError as exc:
            raise MissingScopeContext("MetricsScope requested but not defined!") from exc

    @staticmethod
    def _current_dependencies() -> ScopeDependencies:
        try:
            return _DependenciesScope_Var.get()
        except LookupError as exc:
            raise MissingScopeContext("DependenciesScope requested but not defined!") from exc

    @staticmethod
    def _current_state() -> ScopeState:
        try:
            return _StateScope_Var.get()
        except LookupError as exc:
            raise MissingScopeContext("StateScope requested but not defined!") from exc

    @staticmethod
    def spawn_task[**Args, Result](
        function: Callable[Args, Coroutine[None, None, Result]],
        /,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Task[Result]:
        nested_context: _PartialContext = ctx.nested(function.__name__)

        async def wrapped(*args: Args.args, **kwargs: Args.kwargs) -> Result:
            with nested_context:
                return await function(*args, **kwargs)

        return ctx._current_task_group().create_task(wrapped(*args, **kwargs))

    @staticmethod
    def spawn_subtask[**Args, Result](
        function: Callable[Args, Coroutine[None, None, Result]],
        /,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Task[Result]:
        return ctx._current_task_group().create_task(function(*args, **kwargs))

    @staticmethod
    def cancel() -> None:
        if task := current_task():
            task.cancel()
        else:
            raise RuntimeError("Attempting to cancel context out of asyncio task")

    @staticmethod
    def nested(
        label: str,
        /,
        state: ScopeState | Iterable[ParametrizedData] | None = None,
        metrics: Iterable[Metric] | None = None,
    ) -> _PartialContext:
        nested_state: ScopeState | None
        if isinstance(state, ScopeState):
            nested_state = state
        else:
            nested_state = ctx._current_state().updated(state)

        return _PartialContext(
            task_group=TaskGroup(),
            metrics=ctx._current_metrics().nested(
                label=label,
                metrics=metrics,
            ),
            state=nested_state,
        )

    @staticmethod
    def stream[Element](
        generator: AsyncGenerator[Element, None],
    ) -> AsyncIterator[Element]:
        # TODO: find better solution for streaming without spawning tasks if able
        stream: AsyncStream[Element] = AsyncStream()
        current_metrics: MetricsTrace = ctx._current_metrics()
        current_metrics.enter()  # ensure valid metrics scope closing

        async def iterate() -> None:
            try:
                async for element in generator:
                    stream.send(element)

            except BaseException as exc:
                stream.finish(exception=exc)

            else:
                stream.finish()

            finally:
                current_metrics.exit()

        ctx.spawn_subtask(iterate)
        return stream

    @staticmethod
    def updated(
        *state: ParametrizedData,
    ) -> _PartialContext:
        return _PartialContext(state=ctx._current_state().updated(state))

    @staticmethod
    def id() -> str:
        return ctx._current_metrics().trace_id

    @staticmethod
    def state[State_T: ParametrizedData](
        state: type[State_T],
        /,
    ) -> State_T:
        return ctx._current_state().state(state)

    @staticmethod
    def dependency[Dependency_T: ScopeDependency](
        dependency: type[Dependency_T],
        /,
    ) -> Dependency_T:
        return ctx._current_dependencies().dependency(dependency)

    @staticmethod
    def read[Metric_T: Metric](
        metric: type[Metric_T],
        /,
    ) -> Metric_T | None:
        return ctx._current_metrics().read(metric)

    @staticmethod
    def record(
        *metrics: Metric,
    ) -> None:
        try:
            ctx._current_metrics().record(*metrics)

        # ignoring metrics record when using out of metrics context
        # using default logger as fallback as we already know that we are missing metrics
        except MissingScopeContext as exc:
            logger: Logger = getLogger()
            logger.error("Attempting to record metrics outside of metrics context")
            logger.error(
                exc,
                exc_info=True,
            )

    @staticmethod
    def log_error(
        message: str,
        /,
        *args: Any,
        exception: BaseException | None = None,
    ) -> None:
        try:
            ctx._current_metrics().log_error(
                message,
                *args,
                exception=exception,
            )

        # using default logger as fallback when using out of metrics context
        except MissingScopeContext:
            logger: Logger = getLogger()
            logger.error(
                message,
                *args,
            )
            if exception := exception:
                logger.error(
                    exception,
                    exc_info=True,
                )

    @staticmethod
    def log_warning(
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        try:
            ctx._current_metrics().log_warning(
                message,
                *args,
                exception=exception,
            )

        # using default logger as fallback when using out of metrics context
        except MissingScopeContext:
            logger: Logger = getLogger()
            logger.warning(
                message,
                *args,
            )
            if exception := exception:
                logger.error(
                    exception,
                    exc_info=True,
                )

    @staticmethod
    def log_info(
        message: str,
        /,
        *args: Any,
    ) -> None:
        try:
            ctx._current_metrics().log_info(
                message,
                *args,
            )

        # using default logger as fallback when using out of metrics context
        except MissingScopeContext:
            getLogger().info(
                message,
                *args,
            )

    @staticmethod
    def log_debug(
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        try:
            ctx._current_metrics().log_debug(
                message,
                *args,
                exception=exception,
            )

        # using default logger as fallback when using out of metrics context
        except MissingScopeContext:
            logger: Logger = getLogger()
            logger.debug(
                message,
                *args,
            )
            if exception := exception:
                logger.error(
                    exception,
                    exc_info=True,
                )



import builtins
import types
import typing
from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any, final, get_args, get_origin, overload

from draive.utils import freeze

__all__ = [
    "ParameterPath",
]


class ParameterPathComponent(ABC):
    @abstractmethod
    def path_str(
        self,
        current: str,
    ) -> str: ...

    @abstractmethod
    def resolve(
        self,
        subject: Any,
        /,
    ) -> Any: ...


@final
class ParameterPathAttributeComponent(ParameterPathComponent):
    def __init__[Root, Parameter](
        self,
        root: type[Root],
        parameter: type[Parameter],
        *,
        attribute: str,
    ) -> None:
        root_origin: Any = get_origin(root) or root
        parameter_origin: Any = get_origin(parameter) or parameter

        def resolve(
            subject: Root,
            /,
        ) -> Parameter:
            assert isinstance(subject, root_origin), (  # nosec: B101
                f"ParameterPathComponent used on unexpected root of "
                f"'{type(root)}' instead of '{root}' for '{attribute}'"
            )

            resolved: Any = getattr(subject, attribute)

            assert isinstance(resolved, parameter_origin), (  # nosec: B101
                f"ParameterPathComponent pointing to unexpected value of "
                f"'{type(resolved)}' instead of '{parameter}' for '{attribute}'"
            )
            return resolved

        self._resolve: Callable[[Root], Parameter] = resolve
        self._attribute: str = attribute

        freeze(self)

    def path_str(
        self,
        current: str,
    ) -> str:
        if current:
            return f"{current}.{self._attribute}"
        else:
            return self._attribute

    def resolve(
        self,
        subject: Any,
        /,
    ) -> Any:
        return self._resolve(subject)


@final
class ParameterPathItemComponent(ParameterPathComponent):
    def __init__[Root, Parameter](
        self,
        root: type[Root],
        parameter: type[Parameter],
        *,
        item: Any,
    ) -> None:
        root_origin: Any = get_origin(root) or root
        parameter_origin: Any = get_origin(parameter) or parameter

        def resolve(
            subject: Root,
            /,
        ) -> Parameter:
            assert isinstance(subject, root_origin), (  # nosec: B101
                f"ParameterPathComponent used on unexpected root of "
                f"'{type(root)}' instead of '{root}' for '{item}'"
            )

            resolved: Any = subject.__getitem__(item)  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]

            assert isinstance(resolved, parameter_origin), (  # nosec: B101
                f"ParameterPathComponent pointing to unexpected value of "
                f"'{type(resolved)}' instead of '{parameter}' for '{item}'"
            )
            return resolved

        self._resolve: Callable[[Root], Parameter] = resolve
        self._item: Any = item

    def path_str(
        self,
        current: str | None = None,
    ) -> str:
        if current:
            return f"{current}[{self._item}]"
        else:
            return f"[{self._item}]"

    def resolve(
        self,
        subject: Any,
        /,
    ) -> Any:
        return self._resolve(subject)


@final
class ParameterPath[Root, Parameter]:
    @overload
    def __init__(
        self,
        root: type[Root],
        parameter: type[Root],
    ) -> None: ...

    @overload
    def __init__(
        self,
        root: type[Root],
        parameter: type[Parameter],
        *path: ParameterPathItemComponent,
    ) -> None: ...

    def __init__(
        self,
        root: type[Root],
        parameter: type[Parameter],
        *components: ParameterPathItemComponent,
    ) -> None:
        assert components or root == parameter  # nosec: B101
        self._root: type[Root] = root
        self._parameter: type[Parameter] = parameter
        self._components: tuple[ParameterPathItemComponent, ...] = components

        freeze(self)

    def components(self) -> tuple[str, ...]:
        return tuple(component.path_str() for component in self._components)

    def __str__(self) -> str:
        path: str = ""
        for component in self._components:
            path = component.path_str(path)
        return path

    def __repr__(self) -> str:
        path: str = self._root.__qualname__
        for component in self._components:
            path = component.path_str(path)
        return path

    def __getattr__(
        self,
        name: str,
    ) -> Any:
        try:
            return object.__getattribute__(self, name)

        except (AttributeError, KeyError):
            pass  # continue

        assert not name.startswith(  # nosec: B101
            "_"
        ), f"Accessing private/special parameter paths ({name}) is forbidden"

        try:
            annotation: Any = self._parameter.__annotations__[name]

        except (AttributeError, KeyError) as exc:
            raise AttributeError(name) from exc

        return ParameterPath[Root, Any](
            self._root,
            annotation,
            *[
                *self._components,
                ParameterPathAttributeComponent(
                    root=self._parameter,
                    parameter=annotation,
                    attribute=name,
                ),
            ],
        )

    def __getitem__(
        self,
        key: str | int,
    ) -> Any:
        annotation: Any
        match get_origin(self._parameter) or self._parameter:
            case (
                builtins.list  # pyright: ignore[reportUnknownMemberType]
                | typing.List  # pyright: ignore[reportUnknownMemberType]  # noqa: UP006
            ):
                match get_args(self._parameter):
                    case (element_annotation,) if isinstance(key, int):
                        annotation = element_annotation

                    case other:
                        raise TypeError("Unsupported list type annotation", other)
            case (
                builtins.tuple  # pyright: ignore[reportUnknownMemberType]
                | typing.Tuple  # pyright: ignore[reportUnknownMemberType]  # noqa: UP006
            ):
                match get_args(self._parameter):
                    case (element_annotation, builtins.Ellipsis | types.EllipsisType):
                        annotation = element_annotation

                    case other if isinstance(key, int):
                        annotation = other[key]

                    case other:
                        raise TypeError("Unsupported type annotation", other)
            case (
                builtins.dict  # pyright: ignore[reportUnknownMemberType]
                | typing.Dict  # pyright: ignore[reportUnknownMemberType]  # noqa: UP006
            ):
                match get_args(self._parameter):
                    case (builtins.str, element_annotation) if isinstance(key, str):
                        annotation = element_annotation

                    case (builtins.int, element_annotation) if isinstance(key, int):
                        annotation = element_annotation

                    case other:  # pyright: ignore[reportUnnecessaryComparison]
                        raise TypeError("Unsupported dict type annotation", other)

            case other:
                raise TypeError("Unsupported type annotation", other)

        return ParameterPath[Root, Any](
            self._root,
            annotation,
            *[
                *self._components,
                ParameterPathItemComponent(
                    root=self._parameter,
                    parameter=annotation,
                    item=key,
                ),
            ],
        )

    def __call__(
        self,
        root: Root,
    ) -> Parameter:
        assert isinstance(root, get_origin(self._root) or self._root), (  # nosec: B101
            f"ParameterPath '{self.__repr__()}' used on unexpected root of "
            f"'{type(root)}' instead of '{self._root}'"
        )

        resolved: Any = root
        for component in self._components:
            resolved = component.resolve(resolved)

        assert isinstance(resolved, get_origin(self._parameter) or self._parameter), (  # nosec: B101
            f"ParameterPath '{self.__repr__()}' pointing to unexpected value of "
            f"'{type(resolved)}' instead of '{self._parameter}'"
        )
        return resolved



from abc import ABC, abstractmethod
from asyncio import Lock
from types import TracebackType
from typing import Any, Protocol, Self, final, overload, runtime_checkable
from uuid import UUID, uuid4

from draive.parameters import DataModel, Field, ParameterPath, ParametrizedData, Stateless
from draive.types import Memory, MultimodalContent, MultimodalContentConvertible, frozenlist
from draive.utils import AsyncStream, freeze

__all__ = [
    "AgentBase",
    "AgentInput",
    "AgentOutput",
    "AgentMessage",
    "AgentMessageDraft",
    "AgentCurrent",
    "StatelessAgentCurrent",
    "AgentMessageDraftGroup",
]


class AgentWorkflowStateRead[WorkflowState: ParametrizedData](Protocol):
    def __call__(self) -> WorkflowState: ...


class AgentWorkflowStateUpdate(Protocol):
    def __call__(
        self,
        **parameters: Any,
    ) -> None: ...


@final
class AgentWorkflowStateAccess[WorkflowState: ParametrizedData]:
    def __init__(
        self,
        lock: Lock,
        read: AgentWorkflowStateRead[WorkflowState],
        update: AgentWorkflowStateUpdate,
    ) -> None:
        self._lock: Lock = lock
        self._read: AgentWorkflowStateRead[WorkflowState] = read
        self._update: AgentWorkflowStateUpdate = update

        freeze(self)

    @property
    def current(self) -> WorkflowState:
        assert self._lock.locked(), "Can't access workflow state out of lock!"  # nosec: B101
        return self._read()

    def update(
        self,
        /,
        **parameters: Any,
    ) -> None:
        assert self._lock.locked(), "Can't access workflow state out of lock!"  # nosec: B101
        self._update(**parameters)

    async def __aenter__(self) -> Self:
        await self._lock.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._lock.__aexit__(
            exc_type,
            exc_val,
            exc_tb,
        )


class AgentBase(ABC):
    @property
    @abstractmethod
    def identifier(self) -> UUID: ...

    @abstractmethod
    def prepare_message(
        self,
        content: MultimodalContent | MultimodalContentConvertible,
        *,
        addressee: "AgentBase | None" = None,
    ) -> "AgentMessageDraft": ...


class WorkflowAgentBase[WorkflowState: ParametrizedData, WorkflowResult](AgentBase):
    @abstractmethod
    async def _draive(  # yes, it is the name of the library
        self,
        input: "AgentInput",  # noqa: A002
        workflow: "AgentWorkflowCurrent[WorkflowState, WorkflowResult]",
    ) -> None: ...

    @abstractmethod
    async def start_workflow(
        self,
        input: MultimodalContent | MultimodalContentConvertible,  # noqa: A002
        /,
        state: WorkflowState,
        timeout: float | None = None,
    ) -> WorkflowResult: ...


class AgentMessageDraft(DataModel):
    recipient: AgentBase
    addressee: AgentBase | None
    content: MultimodalContent


class AgentMessageDraftGroup:
    def __init__(
        self,
        messages: AgentMessageDraft,
        /,
        *__messages: AgentMessageDraft,
    ) -> None:
        self.messages: frozenlist[AgentMessageDraft] = (messages, *__messages)

        freeze(self)


class AgentMessage(DataModel):
    identifier: UUID = Field(default_factory=lambda: uuid4())
    sender: AgentBase
    recipient: AgentBase
    addressee: AgentBase | None
    content: MultimodalContent

    @property
    def expects_response(self) -> bool:
        return self.addressee is not None

    def prepare_response(
        self,
        content: MultimodalContent | MultimodalContentConvertible,
        addressee: AgentBase | None = None,
        /,
    ) -> AgentMessageDraft:
        return AgentMessageDraft(
            recipient=self.addressee or self.sender,
            addressee=addressee,
            content=MultimodalContent.of(content),
        )


type AgentInput = AsyncStream[AgentMessage]
type AgentOutput[WorkflowResult] = (
    AgentMessageDraftGroup | AgentMessageDraft | WorkflowResult | None
)


class WorkflowStateAccess[WorkflowState: ParametrizedData](Protocol):
    def __call__[Parameter: ParametrizedData](
        self,
        path: ParameterPath[WorkflowState, Parameter] | Parameter | None,
        /,
    ) -> AgentWorkflowStateAccess[Parameter]: ...


class WorkflowMessageSend(Protocol):
    def __call__(
        self,
        *messages: AgentMessage,
    ) -> None: ...


class WorkflowFinish[WorkflowResult](Protocol):
    def __call__(
        self,
        result: WorkflowResult | BaseException,
        /,
    ) -> None: ...


@final
class AgentWorkflowCurrent[WorkflowState: ParametrizedData, WorkflowResult]:
    def __init__(
        self,
        access: WorkflowStateAccess[WorkflowState],
        send: WorkflowMessageSend,
        finish: WorkflowFinish[WorkflowResult],
    ) -> None:
        self._access: WorkflowStateAccess[WorkflowState] = access
        self._send: WorkflowMessageSend = send
        self._finish: WorkflowFinish[WorkflowResult] = finish

    @overload
    def state(
        self,
        /,
    ) -> AgentWorkflowStateAccess[WorkflowState]: ...

    @overload
    def state[Parameter: ParametrizedData](
        self,
        path: ParameterPath[WorkflowState, Parameter] | Parameter,
        /,
    ) -> AgentWorkflowStateAccess[Parameter]: ...

    def state[Parameter: ParametrizedData](
        self,
        path: ParameterPath[WorkflowState, Parameter] | Parameter | None = None,
        /,
    ) -> AgentWorkflowStateAccess[Parameter]:
        return self._access(path)

    async def update(
        self,
        /,
        **parameters: Any,
    ) -> None:
        async with self.state() as access:
            access.update(**parameters)

    def send(
        self,
        *messages: AgentMessage,
    ) -> None:
        self._send(*messages)

    def finish_with(
        self,
        result: WorkflowResult | BaseException,
        /,
    ) -> None:
        return self._finish(result)


@final
class AgentCurrent[
    AgentState,
    AgentStateScratch,
    WorkflowState: ParametrizedData,
    WorkflowResult,
]:
    def __init__(
        self,
        agent: AgentBase,
        memory: Memory[AgentState, AgentStateScratch],
        workflow: AgentWorkflowCurrent[WorkflowState, WorkflowResult],
    ) -> None:
        self._agent: AgentBase = agent
        self._memory: Memory[AgentState, AgentStateScratch] = memory
        self._workflow: AgentWorkflowCurrent[WorkflowState, WorkflowResult] = workflow

        freeze(self)

    @property
    async def state(self) -> AgentState:
        return await self._memory.recall()

    async def remember(
        self,
        *items: AgentStateScratch,
    ) -> None:
        await self._memory.remember(*items)

    @property
    def workflow(self) -> AgentWorkflowCurrent[WorkflowState, WorkflowResult]:
        return self._workflow

    def send(
        self,
        *messages: AgentMessageDraft,
    ) -> None:
        self._workflow.send(
            *(
                AgentMessage(
                    sender=self._agent,
                    recipient=message.recipient,
                    addressee=message.addressee,
                    content=message.content,
                )
                for message in messages
            )
        )


type StatelessAgentCurrent[
    WorkflowState: ParametrizedData,
    WorkflowResult,
] = AgentCurrent[Stateless, Stateless, WorkflowState, WorkflowResult]


@runtime_checkable
class AgentInvocation[
    AgentState,
    AgentStateScratch,
    WorkflowState: ParametrizedData,
    WorkflowResult,
](Protocol):
    async def __call__(
        self,
        current: AgentCurrent[AgentState, AgentStateScratch, WorkflowState, WorkflowResult],
        message: AgentMessage,
    ) -> AgentOutput[WorkflowResult]: ...



from draive.parameters.data import ParametrizedData

__all__ = [
    "State",
    "Stateless",
]


class State(ParametrizedData):
    def __str__(self) -> str:
        return str(self.as_dict(aliased=False))


class Stateless(State):
    pass



__all__ = [
    "AgentException",
]


class AgentException(Exception):
    def __init__(self, cause: BaseException, *args: object) -> None:
        self.__cause__ = cause
        super().__init__(*args)



from asyncio import Task
from typing import Self, final
from uuid import UUID

from draive.agents.types import AgentInput, AgentMessage, AgentWorkflowCurrent, WorkflowAgentBase
from draive.parameters import ParametrizedData
from draive.scope import ctx
from draive.utils import AsyncStream, freeze

__all__ = [
    "AgentRunner",
]


@final
class AgentRunner:
    @classmethod
    def spawn[
        WorkflowState: ParametrizedData,
        WorkflowResult,
    ](
        cls,
        agent: WorkflowAgentBase[WorkflowState, WorkflowResult],
        /,
        workflow: AgentWorkflowCurrent[WorkflowState, WorkflowResult],
    ) -> Self:
        agent_input: AgentInput = AsyncStream()

        return cls(
            identifier=agent.identifier,
            input=agent_input,
            task=ctx.spawn_subtask(
                agent._draive,  # pyright: ignore[reportPrivateUsage]
                agent_input,
                workflow,
            ),
        )

    def __init__(
        self,
        identifier: UUID,
        input: AgentInput,  # noqa: A002
        task: Task[None],
    ) -> None:
        self.identifier: UUID = identifier
        self._input: AgentInput = input
        self._task: Task[None] = task

        freeze(self)

    def send(
        self,
        message: AgentMessage,
        /,
    ) -> None:
        self._input.send(message)

    def finish(self) -> None:
        if not self._input.finished:
            self._input.finish()

    async def finalize(self) -> None:
        self.finish()

        await self._input.wait()
        await self._task



from abc import ABC, abstractmethod

__all__ = [
    "Memory",
]


class Memory[Recalled, Remembered](ABC):
    @abstractmethod
    async def recall(self) -> Recalled: ...

    @abstractmethod
    async def remember(
        self,
        *items: Remembered,
    ) -> None: ...



from abc import ABC, abstractmethod
from asyncio import gather
from typing import Self, cast, final

from draive.scope.errors import MissingScopeDependency

__all__ = [
    "ScopeDependencies",
    "ScopeDependency",
]


class ScopeDependency(ABC):
    @classmethod
    def interface(cls) -> type:
        return cls

    @classmethod
    @abstractmethod
    def prepare(cls) -> Self: ...

    async def dispose(self) -> None:  # noqa: B027
        pass


@final
class ScopeDependencies:
    def __init__(
        self,
        *dependencies: ScopeDependency | type[ScopeDependency],
    ) -> None:
        self._dependencies: dict[type[ScopeDependency], ScopeDependency] = {}
        for dependency in dependencies:
            if isinstance(dependency, ScopeDependency):
                self._dependencies[type(dependency).interface()] = dependency
            else:
                self._dependencies[dependency.interface()] = dependency.prepare()

    def dependency[Dependency_T: ScopeDependency](
        self,
        dependency: type[Dependency_T],
        /,
    ) -> Dependency_T:
        if dependency in self._dependencies:
            return cast(Dependency_T, self._dependencies[dependency])

        else:
            raise MissingScopeDependency(
                f"{dependency.__qualname__} is not defined!"
                " You have to define it when creating a new context."
            )

    async def dispose(self) -> None:
        await gather(*[dependency.dispose() for dependency in self._dependencies.values()])



from collections.abc import Callable
from inspect import isfunction
from typing import Protocol, cast, final, overload
from uuid import UUID, uuid4

from draive.agents.errors import AgentException
from draive.agents.types import (
    AgentBase,
    AgentCurrent,
    AgentInput,
    AgentInvocation,
    AgentMessage,
    AgentMessageDraft,
    AgentMessageDraftGroup,
    AgentOutput,
    AgentWorkflowCurrent,
    WorkflowAgentBase,
)
from draive.agents.workflow import AgentWorkflow
from draive.helpers import ConstantMemory, VolatileMemory
from draive.parameters import ParametrizedData, Stateless
from draive.scope import ctx
from draive.types import (
    Memory,
    MultimodalContent,
    MultimodalContentConvertible,
)
from draive.utils import freeze, mimic_function

__all__ = [
    "agent",
    "Agent",
]


@final
class Agent[
    AgentState,
    AgentStateScratch,
    WorkflowState: ParametrizedData,
    WorkflowResult,
](WorkflowAgentBase[WorkflowState, WorkflowResult]):
    def __init__(
        self,
        name: str,
        description: str | None,
        memory: Callable[[], Memory[AgentState, AgentStateScratch]],
        invocation: AgentInvocation[
            AgentState,
            AgentStateScratch,
            WorkflowState,
            WorkflowResult,
        ],
    ) -> None:
        self._identifier: UUID = uuid4()
        self.name: str = name
        self.description: str | None = description
        self._memory: Callable[[], Memory[AgentState, AgentStateScratch]] = memory
        self._invocation: AgentInvocation[
            AgentState,
            AgentStateScratch,
            WorkflowState,
            WorkflowResult,
        ] = invocation

        mimic_function(invocation, within=self)
        freeze(self)

    def __eq__(
        self,
        other: object,
    ) -> bool:
        if isinstance(other, Agent):
            return self._identifier == other._identifier
        else:
            return False

    def __hash__(self) -> int:
        return hash(self._identifier)

    def __str__(self) -> str:
        return f"Agent|{self.name}|{self._identifier}"

    async def __call__(
        self,
        current: AgentCurrent[
            AgentState,
            AgentState,
            WorkflowState,
            WorkflowResult,
        ],
        message: AgentMessage,
    ) -> AgentOutput[WorkflowResult]:
        with ctx.nested(
            f"{self.__str__()}|Invocation|{message.identifier}",
            metrics=[message],
        ):
            return await self._invocation(
                current,
                message,
            )

    async def _draive(  # yes, it is the name of the library
        self,
        input: AgentInput,  # noqa: A002
        workflow: AgentWorkflowCurrent[WorkflowState, WorkflowResult],
    ) -> None:
        with ctx.nested(self.__str__()):
            agent_current = AgentCurrent(
                agent=self,
                memory=self._memory(),
                workflow=workflow,
            )
            async for message in input:
                try:
                    match await self(
                        current=agent_current,
                        message=message,
                    ):
                        case None:
                            break  # wait without action

                        case AgentMessageDraftGroup() as messages_group:
                            workflow.send(
                                *(
                                    AgentMessage(
                                        sender=self,
                                        recipient=message.recipient,
                                        addressee=message.addressee,
                                        content=message.content,
                                    )
                                    for message in messages_group.messages
                                )
                            )

                        case AgentMessageDraft() as message:
                            workflow.send(
                                AgentMessage(
                                    sender=self,
                                    recipient=message.recipient,
                                    addressee=message.addressee,
                                    content=message.content,
                                )
                            )

                        case result:
                            workflow.finish_with(result)

                except Exception as exc:
                    workflow.finish_with(AgentException(cause=exc))

    @property
    def identifier(self) -> UUID:
        return self._identifier

    def prepare_message(
        self,
        content: MultimodalContent | MultimodalContentConvertible,
        *,
        addressee: AgentBase | None = None,
    ) -> AgentMessageDraft:
        return AgentMessageDraft(
            recipient=self,
            addressee=addressee,
            content=MultimodalContent.of(content),
        )

    async def start_workflow(
        self,
        input: MultimodalContent | MultimodalContentConvertible,  # noqa: A002
        /,
        state: WorkflowState,
        timeout: float | None = None,
    ) -> WorkflowResult:
        async with ctx.nested(f"{self}|Workflow"):
            workflow: AgentWorkflow[WorkflowState, WorkflowResult] = AgentWorkflow(
                state=state,
                timeout=timeout or 600,  # default timeout is 10 minutes
            )

            workflow.send(
                AgentMessage(
                    sender=self,
                    recipient=self,
                    addressee=None,
                    content=MultimodalContent.of(input),
                )
            )

            return await workflow.execute()


class PartialAgentWrapper[LocalState, LocalStateScratch](Protocol):
    def __call__[WorkflowState: ParametrizedData, WorkflowResult](
        self,
        invocation: AgentInvocation[LocalState, LocalStateScratch, WorkflowState, WorkflowResult],
    ) -> Agent[LocalState, LocalStateScratch, WorkflowState, WorkflowResult]: ...


@overload
def agent[WorkflowState: ParametrizedData, WorkflowResult](
    invocation: AgentInvocation[Stateless, Stateless, WorkflowState, WorkflowResult],
    /,
) -> Agent[Stateless, Stateless, WorkflowState, WorkflowResult]: ...


@overload
def agent[LocalState](
    *,
    name: str | None = None,
    description: str | None = None,
    initial_state: Callable[[], LocalState],
) -> PartialAgentWrapper[LocalState, LocalState]: ...


@overload
def agent[LocalState, LocalStateScratch](
    *,
    name: str | None = None,
    description: str | None = None,
    memory: Memory[LocalState, LocalStateScratch],
) -> PartialAgentWrapper[LocalState, LocalStateScratch]: ...


def agent[
    LocalState,
    LocalStateScratch,
    WorkflowState: ParametrizedData,
    WorkflowResult,
](
    invocation: AgentInvocation[LocalState, LocalStateScratch, WorkflowState, WorkflowResult]
    | None = None,
    *,
    name: str | None = None,
    description: str | None = None,
    initial_state: Callable[[], LocalState] | None = None,
    memory: Memory[LocalState, LocalStateScratch] | None = None,
) -> (
    PartialAgentWrapper[LocalState, LocalStateScratch]
    | Agent[LocalState, LocalStateScratch, WorkflowState, WorkflowResult]
):
    assert initial_state is None or memory is None, "Can't specify both initial state and memory"  # nosec: B101

    def agent_memory() -> Memory[LocalState, LocalStateScratch]:
        if memory is not None:
            return memory

        elif initial_state is not None:
            return cast(Memory[LocalState, LocalStateScratch], VolatileMemory(initial_state()))

        else:
            return cast(Memory[LocalState, LocalStateScratch], ConstantMemory(Stateless()))

    def wrap[WrappedWorkflowState: ParametrizedData, WrappedWorkflowResult](
        invocation: AgentInvocation[
            LocalState, LocalStateScratch, WrappedWorkflowState, WrappedWorkflowResult
        ],
    ) -> Agent[LocalState, LocalStateScratch, WrappedWorkflowState, WrappedWorkflowResult]:
        assert isfunction(invocation), "Agent has to be defined from a function"  # nosec: B101

        return Agent(
            name=name or invocation.__qualname__,
            description=description,
            memory=agent_memory,
            invocation=invocation,
        )

    if invocation := invocation:
        return wrap(invocation)

    else:
        return wrap



from typing import Literal

from draive.parameters import DataModel
from draive.utils import MISSING, Missing

__all__ = [
    "MRSChatConfig",
]


class MRSChatConfig(DataModel):
    model: Literal["Phi-3"] | str = "Phi-3"
    temperature: float = 0.0
    top_p: float | Missing = MISSING
    top_k: int | Missing = MISSING
    max_tokens: int | Missing = MISSING



# from draive.agents.pool import AgentPool, AgentPoolCoordinator
from draive.agents.definition import Agent, agent
from draive.agents.errors import AgentException
from draive.agents.types import (
    AgentCurrent,
    AgentInput,
    AgentMessage,
    AgentMessageDraft,
    AgentMessageDraftGroup,
    AgentOutput,
    AgentWorkflowCurrent,
    AgentWorkflowStateAccess,
    StatelessAgentCurrent,
)
from draive.agents.workflow import AgentWorkflow

__all__ = [
    "agent",
    "Agent",
    "AgentWorkflow",
    "AgentException",
    "AgentMessage",
    "AgentMessage",
    "AgentCurrent",
    "AgentWorkflowCurrent",
    "AgentWorkflowStateAccess",
    "AgentInput",
    "AgentOutput",
    "AgentMessageDraft",
    "StatelessAgentCurrent",
    "AgentMessageDraftGroup",
]



import json
from http import HTTPStatus
from typing import Any, Literal, Self, cast, final

from httpx import AsyncClient, Response

from draive.ollama.config import OllamaChatConfig
from draive.ollama.errors import OllamaException
from draive.ollama.models import (
    ChatCompletionResponse,
    ChatMessage,
)
from draive.parameters import DataModel
from draive.scope import ScopeDependency
from draive.utils import getenv_str, not_missing

__all__ = [
    "OllamaClient",
]


@final
class OllamaClient(ScopeDependency):
    @classmethod
    def prepare(cls) -> Self:
        return cls(
            endpoint=getenv_str("OLLAMA_ENDPOINT", "http://localhost:11434"),
            timeout=90,
        )

    def __init__(
        self,
        endpoint: str,
        timeout: float | None = None,
    ) -> None:
        self._client: AsyncClient = AsyncClient(
            base_url=endpoint,
            timeout=timeout,
        )

    async def chat_completion(
        self,
        *,
        config: OllamaChatConfig,
        messages: list[ChatMessage],
    ) -> ChatCompletionResponse:
        return await self._create_chat_completion(
            messages=messages,
            model=config.model,
            temperature=config.temperature,
            top_k=config.top_k if not_missing(config.top_k) else None,
            top_p=config.top_p if not_missing(config.top_p) else None,
            max_tokens=config.max_tokens if not_missing(config.max_tokens) else None,
            response_format=cast(Literal["text", "json"], config.response_format)
            if not_missing(config.response_format)
            else "text",
            seed=config.seed if not_missing(config.seed) else None,
        )

    async def _create_chat_completion(  # noqa: PLR0913
        self,
        model: str,
        temperature: float,
        top_k: float | None,
        top_p: float | None,
        seed: int | None,
        max_tokens: int | None,
        response_format: Literal["text", "json"],
        messages: list[ChatMessage],
    ) -> ChatCompletionResponse:
        request_body: dict[str, Any] = {
            "model": model,
            "messages": [message.as_dict() for message in messages],
            "options": {
                "temperature": temperature,
            },
            "stream": False,
        }

        if max_tokens:
            request_body["options"]["num_predict"] = max_tokens
        if top_k is not None:
            request_body["options"]["top_k"] = top_k
        if top_p is not None:
            request_body["options"]["top_p"] = top_p
        if seed is not None:
            request_body["options"]["seed"] = seed
        if response_format == "json":
            request_body["format"] = "json"

        return await self._request(
            model=ChatCompletionResponse,
            method="POST",
            url="/api/chat",
            body=request_body,
        )

    async def dispose(self) -> None:
        await self._client.aclose()

    async def _request[Requested: DataModel](  # noqa: PLR0913
        self,
        model: type[Requested],
        method: str,
        url: str,
        query: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        body: DataModel | dict[str, Any] | None = None,
        follow_redirects: bool | None = None,
        timeout: float | None = None,
    ) -> Requested:
        request_headers: dict[str, str]
        if headers:
            request_headers = headers
        else:
            request_headers = {
                "Accept": "application/json",
            }

        body_content: str | None
        match body:
            case None:
                body_content = None

            case body_model if isinstance(body_model, DataModel):
                body_content = body_model.as_json()

            case values:
                body_content = json.dumps(values)

        if body_content:
            request_headers["Content-Type"] = "application/json"

        response: Response
        try:
            response = await self._client.request(
                method=method,
                url=url,
                headers=request_headers,
                params=query,
                content=body_content,
                follow_redirects=follow_redirects or False,
                timeout=timeout,
            )

        except Exception as exc:
            raise OllamaException("Network request failed") from exc

        status: HTTPStatus = HTTPStatus(value=response.status_code)
        if status.is_success:
            try:
                return model.from_json(await response.aread())

            except Exception as exc:
                raise OllamaException("Failed to decode Ollama response", response) from exc

        elif status.is_client_error:
            error_body: bytes = await response.aread()
            raise OllamaException("Ollama request error: %s", error_body.decode("utf-8"))

        else:
            raise OllamaException("Network request failed", response)



from asyncio import (
    AbstractEventLoop,
    CancelledError,
    Future,
    Lock,
    TimerHandle,
    get_running_loop,
)
from typing import Any, cast, final
from uuid import UUID

from draive.agents.runner import AgentRunner
from draive.agents.types import (
    AgentMessage,
    AgentWorkflowCurrent,
    AgentWorkflowStateAccess,
    WorkflowAgentBase,
)
from draive.parameters import ParameterPath, ParametrizedData
from draive.scope import ctx
from draive.utils import AsyncStream, freeze

__all__ = [
    "AgentWorkflow",
]


@final
class AgentWorkflow[WorkflowState: ParametrizedData, WorkflowResult]:
    def __init__(
        self,
        state: WorkflowState,
        timeout: float,
    ) -> None:
        current_state: WorkflowState = state

        def state_read() -> WorkflowState:
            return current_state

        def state_update(**parameters: Any) -> None:
            nonlocal current_state
            current_state = current_state.updated(**parameters)

        # TODO: prepare property based slice access
        self._workflow_state: AgentWorkflowStateAccess[WorkflowState] = AgentWorkflowStateAccess(
            lock=Lock(),
            read=state_read,
            update=state_update,
        )
        self._runners: dict[UUID, AgentRunner] = {}
        self._messages: AsyncStream[AgentMessage] = AsyncStream()
        self._result: Future[WorkflowResult] = Future()
        self._timeout: float = timeout

        self._workflow_current: AgentWorkflowCurrent[WorkflowState, WorkflowResult] = (
            AgentWorkflowCurrent(
                access=self.state,
                send=self.send,
                finish=self.finish_with,
            )
        )

        freeze(self)

    def __del__(self) -> None:
        self.finish_with(CancelledError())

    def state[Parameter: ParametrizedData](
        self,
        path: ParameterPath[WorkflowState, Parameter] | Parameter | None,
        /,
    ) -> AgentWorkflowStateAccess[Parameter]:
        if path is not None:
            # TODO: prepare property based slice access
            raise NotImplementedError("Not implemented yet")

        return cast(AgentWorkflowStateAccess[Parameter], self._workflow_state)

    async def _deliver(
        self,
        message: AgentMessage,
        /,
    ) -> None:
        ctx.log_info(
            "Delivering message [%s] from %s to %s",
            message.identifier,
            message.sender,
            message.recipient,
        )
        if __debug__:
            ctx.log_debug(
                "\nMessage [%s] content: %s",
                message.identifier,
                message.content,
            )

        runner: AgentRunner
        if running := self._runners.get(message.recipient.identifier):
            runner = running

        else:
            runner = AgentRunner.spawn(
                cast(WorkflowAgentBase[WorkflowState, WorkflowResult], message.recipient),
                workflow=self._workflow_current,
            )
            self._runners[message.recipient.identifier] = runner

        runner.send(message)

    def send(
        self,
        *messages: AgentMessage,
    ) -> None:
        if self._result.done():
            return ctx.log_debug("Ignoring messages - workflow finished")

        self._messages.send(*messages)

    def finish_with(
        self,
        result: WorkflowResult | BaseException,
        /,
    ) -> None:
        if self._result.done():
            return  # ignore - already done

        for runner in self._runners.values():
            runner.finish()

        self._messages.finish()

        match result:
            case BaseException() as exception:
                self._result.set_exception(exception)

            case result:
                self._result.set_result(result)

    async def execute(self) -> WorkflowResult:
        if self._result.done():
            raise RuntimeError("AgentWorkflow can be executed only once!")

        loop: AbstractEventLoop = get_running_loop()

        def on_timeout(
            future: Future[WorkflowResult],
        ) -> None:
            if future.done():
                return  # ignore if already finished

            # result future on its completion will ensure that task will complete
            future.set_exception(TimeoutError())

        timeout_handle: TimerHandle = loop.call_later(
            self._timeout,
            on_timeout,
            self._result,
        )

        def on_result(
            future: Future[WorkflowResult],
        ) -> None:
            timeout_handle.cancel()  # at this stage we no longer need timeout to trigger
            self._messages.finish()

        self._result.add_done_callback(on_result)

        # TODO: detect idle state - no agent working
        async for message in self._messages:
            await self._deliver(message)

        return await self._result



from pathlib import Path
from typing import Any, cast

from sentencepiece import SentencePieceProcessor  # pyright: ignore[reportMissingTypeStubs]

from draive.gemini.config import GeminiConfig
from draive.scope import ctx
from draive.utils import cache

__all__ = [
    "gemini_tokenize_text",
]


def gemini_tokenize_text(
    text: str,
    **extra: Any,
) -> list[int]:
    return cast(
        list[int],
        _encoding(model_name=ctx.state(GeminiConfig).model).encode(text),  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]
    )


@cache(limit=4)
def _encoding(model_name: str) -> SentencePieceProcessor:
    model_file: str = _mapping.get(model_name, "gemini_tokenizer.model")
    return SentencePieceProcessor(model_file=str(Path(__file__).parent / "tokens" / model_file))  # pyright: ignore[reportCallIssue]


_mapping: dict[str, str] = {
    "gemini-1.5-flash": "gemini_tokenizer.model",
    "gemini-1.5-pro": "gemini_tokenizer.model",
}



from draive.parameters.annotations import ParameterDefaultFactory
from draive.parameters.basic import BasicValue
from draive.parameters.data import Field, ParametrizedData
from draive.parameters.errors import ParameterValidationContext, ParameterValidationError

# from draive.parameters.definition import ParametersDefinition
from draive.parameters.function import Argument, ParametrizedFunction
from draive.parameters.model import DataModel
from draive.parameters.path import ParameterPath
from draive.parameters.requirement import ParameterRequirement
from draive.parameters.specification import (
    ParameterSpecification,
    ParametersSpecification,
    ToolFunctionSpecification,
    ToolSpecification,
)
from draive.parameters.state import State, Stateless
from draive.parameters.validation import ParameterValidator, ParameterVerifier

__all__ = [
    "Argument",
    "BasicValue",
    "DataModel",
    "Field",
    "ParameterDefaultFactory",
    "ParameterPath",
    "ParameterRequirement",
    "ParameterSpecification",
    "ParametersSpecification",
    "ParameterValidationContext",
    "ParameterValidationContext",
    "ParameterValidationError",
    "ParameterValidator",
    "ParameterVerifier",
    "ParametrizedData",
    "ParametrizedData",
    "ParametrizedFunction",
    "State",
    "Stateless",
    "ToolFunctionSpecification",
    "ToolSpecification",
]
